package types

import (
	"database/sql"
	"database/sql/driver"
	"encoding/json"
	"fmt"
	"time"

	"go.uber.org/zap"
)

var dateLayouts = []string{
	"2006-01-02", "2006-1-2", "2006/01/02", "06-1-2", "6-01-02", "01/02/06", "1/2/06", "2006年01月02日", "06年1月2日",
}

type Date struct {
	time.Time
}

func NewDate(year int, month time.Month, day int) Date {
	return Date{
		Time: time.Date(year, month, day, 0, 0, 0, 0, loc),
	}
}

func NewEmptyDate() Date {
	return Date{
		Time: time.Time{}.In(loc),
	}
}

func MinDate() Date {
	return NewDate(1, 1, 1)
}

func MaxDate() Date {
	return NewDate(9999, 12, 31)
}

func NowDate() Date {
	return Now().Date()
}

func ParseDate(t string) (Date, error) {
	if len(t) == 0 {
		return NewEmptyDate(), fmt.Errorf("不能解析空白的日期时间。")
	}
	for _, layout := range dateLayouts {
		d, err := time.ParseInLocation(layout, t, loc)
		if err == nil {
			return Date{
				Time: d,
			}, nil
		}
	}
	return NewEmptyDate(), fmt.Errorf("无法解析给定的日期，格式不正确。")
}

func ParseDatep(t string) (*Date, error) {
	if len(t) == 0 {
		return nil, fmt.Errorf("不能解析空白的日期时间。")
	}
	for _, layout := range dateLayouts {
		d, err := time.ParseInLocation(layout, t, loc)
		if err == nil {
			return &Date{
				Time: d,
			}, nil
		}
	}
	return nil, fmt.Errorf("无法解析给定的日期，格式不正确。")
}

func ParseDateWithDefault(t string, defaultDate Date) Date {
	if len(t) == 0 {
		return defaultDate
	}
	d, err := ParseDate(t)
	if err != nil {
		return defaultDate
	}
	return d
}

var _ driver.Valuer = (*Date)(nil)

func (dt Date) Value() (driver.Value, error) {
	return dt.In(loc).Format("2006-01-02"), nil
}

var _ sql.Scanner = (*Date)(nil)

func (d *Date) Scan(src interface{}) (err error) {
	switch src := src.(type) {
	case time.Time:
		d.Time = src
	case string:
		t, err := time.ParseInLocation("2006-01-02", src, loc)
		if err != nil {
			return err
		}
		*d = Date{Time: t}
	case []byte:
		d.Time, err = time.ParseInLocation("2006-01-02", string(src), loc)
		return err
	case nil:
		d = nil
	default:
		return fmt.Errorf("该数据类型不支持解析到日期: %T", src)
	}
	return nil
}

var _ json.Marshaler = (*Date)(nil)

func (d Date) MarshalJSON() ([]byte, error) {
	return json.Marshal(d.Format("2006-01-02"))
}

var _ json.Unmarshaler = (*Date)(nil)

func (d *Date) UnmarshalJSON(data []byte) error {
	var str string
	if err := json.Unmarshal(data, &str); err != nil {
		return fmt.Errorf("不能解析指定的日期时间值: %w", err)
	}
	t, err := time.ParseInLocation("2006-01-02", str, loc)
	d.Time = t
	return err
}

func (d Date) DifferenceInMonth(d2 *Date) int {
	var differYear, differMonth int
	differYear = d.Year() - d2.Year()
	differMonth = int(d.Month() - d2.Month())
	return differYear*12 + differMonth
}

func (d Date) IsNextMonth(d2 *Date) bool {
	return d.DifferenceInMonth(d2) == 1
}

func (d Date) ToBeginningOfDate() DateTime {
	return FromTime(time.Date(d.Year(), d.Month(), d.Day(), 0, 0, 0, 0, loc))
}

func (d Date) ToEndingOfDate() DateTime {
	return FromTime(time.Date(d.Year(), d.Month(), d.Day(), 23, 59, 59, 999999, loc))
}

func (d Date) IsEmpty() bool {
	return d.Time.IsZero()
}

func (d *Date) Parse(s string) error {
	t, err := ParseDate(s)
	if err != nil {
		return err
	}
	d.Time = t.Time
	return nil
}

func (d Date) ToString() string {
	return d.Time.Format("2006-01-02")
}

func (d Date) ToDateTime() DateTime {
	return FromTime(d.Time)
}

func (d Date) Log(fieldName string) zap.Field {
	return zap.String(fieldName, d.ToString())
}
